<html>
<head>
    <title>Datastore viewer for emulator</title>
</head>
<body>
<h1>Datastore viewer for emulator</h1>

<form>
    <label for="project_name">Project Name</label>
    <input name="project_name" placeholder="Project Name" value="{{ project_name }}" />
    <input type="submit" value="view">
</form>

</body>
</html>
